from pypants.generators.base import PythonPackageGenerator


class LibraryPackageGenerator(PythonPackageGenerator):
    """Generator for a Python library package"""

    PACKAGE_TYPE = "library"
    TOP_LEVEL_DIR = "packages"



"""Contains PY2SFNProjectPackage class"""
import ast
import logging
from pathlib import Path
from typing import Dict

from .base import BuildTarget
from .python_package import PythonPackage

logger = logging.getLogger(__name__)


class PY2SFNProjectPackage(PythonPackage):
    """Represents a py2sfn project build target in Pants"""

    def set_dependencies(self, targets: Dict[str, BuildTarget]) -> None:
        """Compute and set the collection of dependencies as a batch.

        We're overriding the base class method because the generic py2sfn project build
        target only needs to depend on the task packages.

        Args:
            targets: map of package name to instance of BuildTarget

        """
        logger.debug(f"Gathering dependencies for {self.key}")
        for task_package_path in Path(self.package_path).glob("tasks/*/src/*"):
            package_name = task_package_path.name
            if package_name in targets:
                self.dependencies.add(targets[package_name])
                logger.debug(
                    f"Added dependency for existing target: {targets[package_name]}"
                )
        logger.debug(f"Dependencies for {self.key}: {self.dependencies}")

    def add_dependency(
        self, targets: Dict[str, BuildTarget], package_name: str
    ) -> None:
        """Add a new dependency to the set

        We're overriding the base class method because the generic py2sfn project build
        target only needs to depend on the task packages. Therefore, we don't want to
        process any nested Python modules.
        """
        pass

    def _generate_generic_target_node(self) -> ast.Expr:
        """Generate an AST node for a generic ``target`` Pants target"""
        node = ast.Expr(
            value=ast.Call(
                func=ast.Name(id="target"),
                args=[],
                keywords=[self._get_dependencies_keyword(), self._tags_keyword],
            )
        )
        return node

    def generate_build_file_ast_node(self) -> ast.Module:
        """Generate a Pants BUILD file as an AST module node"""
        node = ast.Module(body=[self._generate_generic_target_node()])
        return node



"""Contains PythonRequirement class"""
from ..config import PROJECT_CONFIG  # noqa: I100
from .base import BuildTarget  # noqa: I100

# Parent directory name containing 3rdparty requirements.txt
REQUIREMENTS_DIR = PROJECT_CONFIG.third_party_requirements_path.parent


class PythonRequirement(BuildTarget):
    """Represents a Python requirement build target in Pants"""

    def __init__(self, package_name: str) -> None:
        self.package_name = package_name
        self.key = f"{REQUIREMENTS_DIR}:{self.package_name}"



"""Contains exception classes"""


class CircularImportsFound(Exception):
    """Raised when circular imports were found processing build targets"""

    pass


class NoBuildDependenciesFound(Exception):
    """Raised when the dependencies argument could not be found in a BUILD file AST"""

    pass


class DistributionError(Exception):
    """Raised when a Python distribution can't be parsed"""

    pass


class UnsupportedDistributionFormat(DistributionError):
    """Raised when the distribution file is something other than a .whl or .tar.gz"""

    pass


class MissingDistributionMetadataFile(DistributionError):
    """Raised when there is a missing metadata file in a .whl or .tar.gz"""

    pass


class NoProjectName(Exception):
    """Raised when no project name could be parsed from distribution metadata"""

    pass


class InvalidTopLevelFile(Exception):
    """Raised when a top_level.txt file could not be interpreted"""

    pass


class MultipleSourcePackagesFound(Exception):
    """Raised when more than one package is found in src/"""

    pass


class DuplicateTarget(Exception):
    """Raised when attempting to register a target that has an existing key"""

    pass


class NoConsoleScriptFound(Exception):
    """Raised when no console_scripts are found in setup.py for a binary target"""

    pass


class NoTargetFound(Exception):
    """Raised when no target could be found in the registered graph of targets"""

    pass



"""Contains base PythonPackage class"""
import ast
import logging
import os
import re
from typing import Dict, List, Optional, Set

import astor

from ..config import Config, PROJECT_CONFIG  # noqa
from ..requirements import THIRD_PARTY_IMPORT_MAP  # noqa
from .base import BuildTarget  # noqa
from .requirement import PythonRequirement  # noqa

logger = logging.getLogger(__name__)


COOKIECUTTER_PATH_PATTERN = re.compile(r"\{\{cookiecutter")


class PythonPackage(BuildTarget):
    """Represents a Python package (e.g. lib or test) build target in Pants"""

    def __init__(
        self,
        target_type: str,
        build_template: str,
        top_dir_name: str,
        package_dir_name: str,
        build_dir: str,
        package_path: str,
        package_name: str,
        rendered_package_name: str = None,
        extra_tags: Set = None,
        config: Config = None,
    ) -> None:
        """
        Args:
            target_type: Target type. One of: code, tests
            build_template: Key of the build template, e.g. lib
            top_dir_name: Top directory name, e.g. lib
            package_dir_name: Package directory name, e.g. python_core
            build_dir: Directory where the BUILD file will be stored,
                e.g. lib/python_core/src
            package_path: Path to the package code,
                e.g. lib/python_core/src/ns_python_core
            package_name: Package name, e.g. ns_python_core
            rendered_package_name: Package name to render into the BUILD file. This can
                be set if there's a conflict between the package name and the build
                directory. e.g. ns_python_core_lib
            extra_tags: Set of extra tags to include in the BUILD file in addition to
                the standard set
            config: Configuration for this package, potentially loaded from a
                .pants.cfg file. If not provided it will be loaded.

        """
        self.target_type = target_type
        self.build_template = build_template
        self.top_dir_name = top_dir_name
        self.package_dir_name = package_dir_name
        self.package_dir_path = PROJECT_CONFIG.config_dir_path.joinpath(
            self.top_dir_name, self.package_dir_name
        )
        self.build_dir = build_dir
        self.build_file = os.path.join(self.build_dir, "BUILD")
        self.package_path = package_path
        self.package_name = package_name
        self.rendered_package_name = (
            rendered_package_name
            if rendered_package_name is not None
            else self.package_name
        )

        # This string will be used as an identifier for the build target
        self.key = self.build_dir

        # Set of dependencies to be gathered later
        self.dependencies = set()

        self.config = config or Config(self.package_dir_path)

        tags = {"python", self.target_type, self.top_dir_name}
        if extra_tags is not None:
            tags.update(extra_tags)
        tags.update(self.config.extra_tags)
        self.tags = tags

    def set_dependencies(self, targets: Dict[str, BuildTarget]) -> None:
        """Compute and set the collection of dependencies as a batch.

        This implementation just reinitializes the dependencies attribute. A
        child class may choose to implement this method and skip the ``add_dependency``
        method.

        Args:
            targets: map of package name to instance of BuildTarget

        """
        self.dependencies = set()

    def add_dependency(
        self, targets: Dict[str, BuildTarget], package_name: str
    ) -> None:
        """Add a new dependency to the set

        A child class may choose to implement ``set_dependencies`` and skip this method.

        Args:
            targets: map of package name to instance of BuildTarget
            package_name: dependency package name, as parsed from an import statement

        """
        logger.debug(f"Processing {package_name}")
        if package_name in targets and package_name != self.package_name:
            self.dependencies.add(targets[package_name])
            logger.debug(
                f"Added dependency for existing target: {targets[package_name]}"
            )
        elif package_name in THIRD_PARTY_IMPORT_MAP:
            self.dependencies.add(
                PythonRequirement(THIRD_PARTY_IMPORT_MAP[package_name])
            )
            logger.debug(
                f"Added third-party dependency: {THIRD_PARTY_IMPORT_MAP[package_name]}"
            )

    def _parse_build_ast(self) -> Optional[ast.Module]:
        """Parse AST of BUILD file if it exists

        Returns:
            AST of the BUILD file or ``NoneType`` if the file doesn't exist

        """
        try:
            return astor.code_to_ast.parse_file(self.build_file)
        except FileNotFoundError:
            logger.error(f"Could not find {self.build_file}")
            return None

    def generate_build_file(self) -> ast.Module:
        """Generate a Pants BUILD file for the build target

        Returns:
            AST module node representing the BUILD file

        """
        if not self.config.generate_build_file:
            logger.debug(f"Not writing BUILD file to {self.build_file}")
            return

        return self.generate_build_file_ast_node()

    def generate_build_file_ast_node(self) -> ast.Module:
        """Generate a Pants BUILD file as an AST module node"""
        raise NotImplementedError("generate_build_file_ast_node")

    def _get_dependencies_keyword(
        self, include_extra_dependencies: bool = True
    ) -> ast.keyword:
        """Build the AST node representing the dependencies keyword argument

        Args:
            include_extra_dependencies: Flag for whether the extra dependencies
                specified in the config file should be included

        """
        dependencies = set(
            dependency.dependency_target for dependency in self.dependencies
        )
        if include_extra_dependencies:
            dependencies.update(self.config.extra_dependencies)
        dependencies = sorted(list(dependencies), key=lambda d: str(d).lower())
        return ast.keyword(
            arg="dependencies", value=ast.List(elts=[ast.Str(d) for d in dependencies])
        )

    @property
    def _tags_keyword(self) -> ast.keyword:
        """AST node representing the tags keyword argument"""
        return ast.keyword(
            arg="tags", value=ast.Set(elts=[ast.Str(t) for t in sorted(self.tags)])
        )

    def _generate_python_library_ast_node(
        self,
        name: Optional[str] = None,
        globs_path: str = "**/*",
        include_extra_dependencies: bool = True,
    ) -> ast.Expr:
        """Generate an AST node for a python_library Pants target

        Args:
            name: Name of the library target. If not provided, no name will be set.
            globs_path: File globs to include in the target. Defaults to everything.
            include_extra_dependencies: Flag for whether the extra dependencies
                specified in the config file should be included. This should usually be
                True except when generating a test library target.

        Returns:
            AST expression node

        """
        keywords = [
            self._get_dependencies_keyword(
                include_extra_dependencies=include_extra_dependencies
            ),
            ast.keyword(
                arg="sources",
                value=ast.Call(
                    func=ast.Name(id="globs"), args=[ast.Str(globs_path)], keywords=[]
                ),
            ),
            self._tags_keyword,
        ]
        if name is not None:
            keywords.insert(0, ast.keyword(arg="name", value=ast.Str(name)))
        node = ast.Expr(
            value=ast.Call(
                func=ast.Name(id="python_library"), args=[], keywords=keywords
            )
        )
        return node

    def _generate_python_binary_wrapper_node(
        self,
        name: str,
        entry_point: Optional[str] = None,
        dependencies: Optional[List[str]] = None,
    ) -> ast.Expr:
        """Generate an AST node for a python_binary Pants target with an entry point

        See: https://www.pantsbuild.org/python_readme.html#python_binary-entry_point

        Args:
            name: Name of the binary, e.g. unittest
            entry_point: Entrypoint module in package, e.g. unittest. Defaults to the
                name
            dependencies: List of targets for this wrapper to depend on

        Returns:
            AST expression node

        """
        if entry_point is None:
            entry_point = name

        node = ast.Expr(
            value=ast.Call(
                func=ast.Name(id="python_binary"),
                args=[],
                keywords=[
                    ast.keyword(arg="name", value=ast.Str(name)),
                    ast.keyword(arg="entry_point", value=ast.Str(entry_point)),
                    ast.keyword(
                        arg="dependencies",
                        value=ast.List(elts=[ast.Str(d) for d in dependencies]),
                    ),
                ],
            )
        )
        return node



"""{{ cookiecutter.title }}"""

import logging
import os

import click

logger = logging.getLogger("{{ cookiecutter.binary_name }}")


@click.group()
@click.option(
    "--log-level",
    type=click.Choice(["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]),
    default="INFO",
)
def cli(log_level: str) -> None:
    logger.setLevel(getattr(logging, log_level))


@cli.command()
@click.option("--name", help="Your name", default=lambda: os.environ.get("USER"))
def welcome(name: str) -> None:
    """Welcome!"""
    logger.info("Welcoming...")
    click.echo(f"Welcome! {name}")


if __name__ == "__main__":
    cli()



"""Contains AlembicMigrationPackage class"""
import ast

from .python_package import PythonPackage


class AlembicMigrationPackage(PythonPackage):
    """Represents an Alembic database migration bundle target in Pants"""

    def _generate_bundle_item(self, glob: str) -> ast.Call:
        """Generate an AST node for a bundle item for use in a python_app target"""
        return ast.Call(
            func=ast.Name(id="bundle"),
            args=[],
            keywords=[
                ast.keyword(
                    arg="fileset",
                    value=ast.Call(
                        func=ast.Name(id="globs"), args=[ast.Str(glob)], keywords=[]
                    ),
                )
            ],
        )

    def _generate_migrations_python_app_ast_node(self) -> ast.Expr:
        """Generate an AST node for a python_app Pants target that bundles migrations"""
        node = ast.Expr(
            value=ast.Call(
                func=ast.Name(id="python_app"),
                args=[],
                keywords=[
                    ast.keyword(arg="name", value=ast.Str(self.package_name)),
                    ast.keyword(arg="archive", value=ast.Str("tar")),
                    ast.keyword(arg="binary", value=ast.Str(":alembic")),
                    ast.keyword(
                        arg="bundles",
                        value=ast.List(
                            elts=[
                                self._generate_bundle_item(glob)
                                for glob in ["alembic.ini", "env.py", "versions/*.py"]
                            ]
                        ),
                    ),
                    self._tags_keyword,
                ],
            )
        )
        return node

    def generate_build_file_ast_node(self) -> ast.Module:
        """Generate a Pants BUILD file as an AST module node"""
        node = ast.Module(
            body=[
                self._generate_python_library_ast_node(name="lib"),
                self._generate_python_binary_wrapper_node(
                    "alembic", entry_point="alembic.config", dependencies=[":lib"]
                ),
                self._generate_migrations_python_app_ast_node(),
            ]
        )
        return node



"""Contains PythonLibraryPackage class"""
import ast

from .python_package import PythonPackage


class PythonLibraryPackage(PythonPackage):
    """Represents a Python library package build target in Pants

    A library is a package that is meant to be imported by other packages. It is not
    meant to be run as a script.
    """

    def generate_build_file_ast_node(self) -> ast.Module:
        """Generate a Pants BUILD file as an AST module node"""
        node = ast.Module(
            body=[
                self._generate_python_library_ast_node(
                    globs_path=f"{self.package_name}/**/*"
                )
            ]
        )
        return node



import os

import setuptools

setuptools.setup(
    name="{{ cookiecutter.package_name }}",
    version="0.0.1",
    package_dir={"": "src"},
    packages=setuptools.find_packages("src"),
    provides=setuptools.find_packages("src"),
    entry_points={
        "console_scripts": [
            "{{ cookiecutter.binary_name }} = {{ cookiecutter.package_name }}.cli:cli"
        ]
    },
)



"""Contains functions for working with requirements.txt"""

from copy import deepcopy
import glob
import json
import logging
import os
import subprocess
import sys
import tarfile
import tempfile
from typing import Dict, List, Tuple, TYPE_CHECKING
import zipfile

import pkg_resources

from .config import PROJECT_CONFIG
from .exceptions import (
    DistributionError,
    InvalidTopLevelFile,
    MissingDistributionMetadataFile,
    NoProjectName,
    UnsupportedDistributionFormat,
)

if TYPE_CHECKING:
    import io

logger = logging.getLogger(__name__)


# This file contains a map from the import name found in *.py to the library's project
# name (i.e. requirements.txt name)
try:
    with PROJECT_CONFIG.third_party_import_map_path.open() as f:
        THIRD_PARTY_IMPORT_MAP = json.load(f)
except FileNotFoundError:
    THIRD_PARTY_IMPORT_MAP = {}

# Top-level Python package distribution names to ignore when generating the third-party
# import map
IGNORE_TOP_LEVEL_IMPORTS = {"testing", "tests", "test"}


def _requirement_name(requirement: pkg_resources.Requirement) -> str:
    """Build a full requirement project name that includes extras"""
    name = requirement.name
    if len(requirement.extras) > 1:
        raise NotImplementedError(
            f"Multiple requirement extras are not supported: {requirement.extras}"
        )
    elif len(requirement.extras) == 1:
        name = f"{name}[{requirement.extras[0]}]"

    return name


def _parse_top_level_imports_from_file(
    distribution_path, f: "io.TextIOWrapper"
) -> List[str]:
    """Get the top-level import names from a top_level.txt file

    Args:
        distribution_path: Path to the .whl or .tar.gz file containing package metadata
        f: File handle

    Returns:
        Top-level import names for the distribution (case-sensitive).
        e.g. ``["auth0"]`` for a project name of ``auth0-python``.

    Raises:
        :py:exc:`.InvalidTopLevelFile` when no top level import names could be found

    """
    imports = [
        line
        for line in f.read().decode().split("\n")
        if len(line.strip()) > 0
        and line.strip() not in IGNORE_TOP_LEVEL_IMPORTS
        and not line.strip().startswith("_")
    ]
    if len(imports) == 0:
        raise InvalidTopLevelFile(
            f"No top-level imports found in distribution package for {distribution_path}"
        )

    return imports


def _get_project_name_from_metadata_file(
    distribution_path, f: "io.TextIOWrapper"
) -> str:
    """Get the project name from a METADATA or PKG-INFO file

    Args:
        distribution_path: Path to the .whl or .tar.gz file containing package metadata
        f: File handle

    Returns:
        Project name for the distribution (case-sensitive)

    Raises:
        :py:exc:`.NoProjectName` when no project name could be parsed from distribution
            metadata

    """
    for line in f.readlines():
        if line.decode().startswith("Name:"):
            project_name = line.decode().split(":")[1].strip()
            return project_name

    raise NoProjectName(f"No project name found in {distribution_path}")


def _get_package_names_from_wheel(distribution_path: str) -> Tuple[List[str], str]:
    """Get the top-level import name and project name for a given .whl distribution

    Args:
        distribution_path: Path to the .whl file containing package metadata

    Returns:
        Tuple of ``(top_level_imports, project_name)``

    Raises:
        :py:exc:`.MissingDistributionMetadataFile`

    """
    zf = zipfile.ZipFile(distribution_path)
    top_level_imports = None
    project_name = None
    for name in zf.namelist():
        if os.path.basename(name) == "top_level.txt":
            with zf.open(name) as f:
                top_level_imports = _parse_top_level_imports_from_file(
                    distribution_path, f
                )
        elif name.endswith(".dist-info/METADATA"):
            with zf.open(name) as f:
                project_name = _get_project_name_from_metadata_file(
                    distribution_path, f
                )

        if top_level_imports is not None and project_name is not None:
            break

    if project_name is None:
        raise MissingDistributionMetadataFile(
            f"Could not find METADATA in {distribution_path}"
        )
    elif top_level_imports is None:
        top_level_imports = [project_name]
        logger.debug(
            f"Could not find top_level.txt in {distribution_path}."
            f" Defaulting top level import list to {top_level_imports}"
        )

    return top_level_imports, project_name


def _get_package_names_from_tar(distribution_path: str) -> Tuple[List[str], str]:
    """Get the top-level import name and project name for a given .tar.gz distribution

    Args:
        distribution_path: Path to the .tar.gz file containing package metadata

    Returns:
        Tuple of ``(top_level_imports, project_name)``

    Raises:
        :py:exc:`.MissingDistributionMetadataFile`

    """
    top_level_imports = None
    project_name = None
    with tarfile.open(distribution_path) as tar:
        for tarinfo in tar:
            if tarinfo.isreg():
                if os.path.basename(tarinfo.name) == "top_level.txt":
                    f = tar.extractfile(tarinfo)
                    top_level_imports = _parse_top_level_imports_from_file(
                        distribution_path, f
                    )
                elif os.path.basename(tarinfo.name) == "PKG-INFO":
                    f = tar.extractfile(tarinfo)
                    project_name = _get_project_name_from_metadata_file(
                        distribution_path, f
                    )
                elif tarinfo.name.endswith(".egg-info/PKG-INFO"):
                    f = tar.extractfile(tarinfo)
                    project_name = _get_project_name_from_metadata_file(
                        distribution_path, f
                    )

            if top_level_imports is not None and project_name is not None:
                break

    if project_name is None:
        raise MissingDistributionMetadataFile(
            f"Could not find PKG-INFO in {distribution_path}"
        )
    elif top_level_imports is None:
        top_level_imports = [project_name]
        logger.debug(
            f"Could not find top_level.txt in {distribution_path}."
            f" Defaulting top level import list to {top_level_imports}"
        )

    return top_level_imports, project_name


def _get_package_names_from_zip(distribution_path: str) -> Tuple[List[str], str]:
    """Get the top-level import name and project name for a given .zip distribution

    Args:
        distribution_path: Path to the .zip file containing package metadata

    Returns:
        Tuple of ``(top_level_imports, project_name)``

    Raises:
        :py:exc:`.MissingDistributionMetadataFile`

    """
    zf = zipfile.ZipFile(distribution_path)
    top_level_imports = None
    project_name = None
    for name in zf.namelist():
        if os.path.basename(name) == "top_level.txt":
            with zf.open(name) as f:
                top_level_imports = _parse_top_level_imports_from_file(
                    distribution_path, f
                )
        elif os.path.basename(name) == "PKG-INFO":
            with zf.open(name) as f:
                project_name = _get_project_name_from_metadata_file(
                    distribution_path, f
                )

        if top_level_imports is not None and project_name is not None:
            break

    if project_name is None:
        raise MissingDistributionMetadataFile(
            f"Could not find PKG-INFO in {distribution_path}"
        )
    elif top_level_imports is None:
        top_level_imports = [project_name]
        logger.debug(
            f"Could not find top_level.txt in {distribution_path}."
            f" Defaulting top level import list to {top_level_imports}"
        )

    return top_level_imports, project_name


def _get_package_names_for_distribution(
    distribution_path: str
) -> Tuple[List[str], str]:
    """Get the top-level import names and project name for a given distribution file path

    Args:
        distribution_path: Path to the .whl, .tar.gz, or .zip file containing package
            metadata

    Returns:
        Tuple of ``(top_level_imports, project_name)``

    Raises:
        :py:exc:`.UnsupportedDistributionFormat`

    """
    if distribution_path.endswith(".whl"):
        return _get_package_names_from_wheel(distribution_path)
    elif distribution_path.endswith(".tar.gz") or distribution_path.endswith(
        ".tar.bz2"
    ):
        return _get_package_names_from_tar(distribution_path)
    elif distribution_path.endswith(".zip"):
        return _get_package_names_from_zip(distribution_path)

    raise UnsupportedDistributionFormat(
        f"Unsupported Python distribution file: {distribution_path}"
    )


def _get_import_map_for_requirements(
    requirement_specifiers: List[str],
) -> Tuple[List[Exception], Dict[str, str]]:
    """Get a map of top-level import name to project name for a list of requirements

    Args:
        requirement_specifiers: List of requirement specifiers like ``requests==1.2.3``

    Returns:
        Tuple with items:
        * List of exception objects, if any
        * Map of the top-level import name to the project name

    """
    import_map = {}
    with tempfile.TemporaryDirectory() as tmpdirname:
        logger.debug(f"Downloading Python packages to {tmpdirname}")
        # Download the wheels and tarballs for the requirements into a temp directory
        subprocess.run(
            [
                sys.executable,
                "-m",
                "pip",
                "download",
                "-d",
                tmpdirname,
                "--no-deps",
                *requirement_specifiers,
            ],
            check=True,
        )
        distribution_paths = (
            list(glob.glob(f"{tmpdirname}/*.whl"))
            + list(glob.glob(f"{tmpdirname}/*.tar.gz"))
            + list(glob.glob(f"{tmpdirname}/*.tar.bz2"))
            + list(glob.glob(f"{tmpdirname}/*.zip"))
        )
        logger.debug(f"Found distribution paths: {distribution_paths}")
        # Process each downloaded distribution, extracting the top-level import name
        # and the project name to add to the map
        exceptions = []
        for distribution_path in distribution_paths:
            try:
                top_level_imports, project_name = _get_package_names_for_distribution(
                    distribution_path
                )
            except Exception as e:
                exceptions.append(e)
            else:
                for top_level_import in top_level_imports:
                    logger.debug(
                        f"Found map of {top_level_import} => {project_name} for"
                        f" distribution {distribution_path}"
                    )
                    import_map[top_level_import] = project_name

    return exceptions, import_map


def update_third_party_import_map() -> None:
    """Update the third-party import map file.

    This will determine which keys should be added or dropped based on the
    requirements.txt file. Therefore, :py:func:`generate_3rdparty_requirements` should
    be run prior to this function.

    """
    logger.info("Updating third-party import map")

    with PROJECT_CONFIG.third_party_requirements_path.open() as f:
        requirements = list(pkg_resources.parse_requirements(f.read()))

    existing_project_names = set(THIRD_PARTY_IMPORT_MAP.values())
    current_project_names = set(req.project_name for req in requirements)

    # New requirements for which we need to go resolve an import name
    new_project_names = current_project_names - existing_project_names
    logger.info(
        f"Found {len(new_project_names)} new requirement(s) to resolve: {new_project_names}"
    )

    # Old requirements that should be removed from the map
    old_project_names = existing_project_names - current_project_names
    logger.info(
        f"Found {len(old_project_names)} requirement(s) to remove: {old_project_names}"
    )

    if len(new_project_names) == 0 and len(old_project_names) == 0:
        logger.info("No updates needed for third-party import map")
        return

    # Create a copy of the import map that we'll mutate
    import_map = deepcopy(THIRD_PARTY_IMPORT_MAP)
    import_map_inverse = {v: k for k, v in THIRD_PARTY_IMPORT_MAP.items()}

    # Remove old requirements
    for project_name in old_project_names:
        del import_map[import_map_inverse[project_name]]

    new_requirement_specifiers = [
        str(req) for req in requirements if req.project_name in new_project_names
    ]
    import_map_exceptions = []
    if len(new_requirement_specifiers) > 0:
        import_map_exceptions, import_map_updates = _get_import_map_for_requirements(
            new_requirement_specifiers
        )
        import_map.update(import_map_updates)

    with PROJECT_CONFIG.third_party_import_map_path.open("w") as f:
        json.dump(import_map, f, sort_keys=True, indent=2)

    logger.info("Finished updating third-party import map")

    if len(import_map_exceptions) > 0:
        messages = "\n" + "\n".join([str(e) for e in import_map_exceptions])
        raise DistributionError(
            f"Failed to process all Python distributions: {messages}"
        )



"""Contains PythonBehaveTestPackage class"""
import ast

from .test import PythonTestPackage


class PythonBehaveTestPackage(PythonTestPackage):
    """Represents a Python test package build target in Pants that uses behave"""

    def _generate_python_binary_behave_wrapper_node(self) -> ast.Expr:
        """Generate an AST node for a python_binary Pants target that wraps behave"""
        node = ast.Expr(
            value=ast.Call(
                func=ast.Name(id="python_binary"),
                args=[],
                keywords=[
                    ast.keyword(arg="source", value=ast.Str("behave_cli.py")),
                    ast.keyword(
                        arg="dependencies", value=ast.List(elts=[ast.Str(":lib")])
                    ),
                    self._tags_keyword,
                ],
            )
        )
        return node

    def generate_build_file_ast_node(self) -> ast.Module:
        """Generate a Pants BUILD file as an AST module node"""
        node = ast.Module(
            body=[
                self._generate_python_library_ast_node(name="lib"),
                self._generate_python_binary_behave_wrapper_node(),
            ]
        )
        return node



"""Contains package processor class"""
import copy
import importlib.util
import logging
from multiprocessing import Pool
import os
from pathlib import Path
from typing import Dict, List, Tuple

import networkx as nx

from .build_targets import (
    BUILD_TARGET_MAP,
    PY2SFNProjectPackage,
    PythonLibraryPackage,
    PythonPackage,
    PythonRequirement,
    PythonTestPackage,
)
from .config import Config, PROJECT_CONFIG
from .exceptions import (
    CircularImportsFound,
    DuplicateTarget,
    MultipleSourcePackagesFound,
    NoTargetFound,
)
from .util import gather_dependencies_from_module, write_build_file

logger = logging.getLogger(__name__)


class PackageProcessor:
    """Class with methods for processing internal Python packages in order to:

    * Check for circular imports
    * Generate Pants BUILD files
    * Determine package dependencies
    """

    def __init__(self) -> None:
        # Map of target key to instance of a Python package build target
        self._targets: Dict[str, PythonPackage] = {}
        # Graph of targets
        self._target_graph: nx.DiGraph = None

    def register_packages(self) -> None:
        """Register targets and their dependencies.

        This should be called before performing an action with the packages.
        """
        self._register_task_targets_code()
        self._register_task_targets_py2sfn_projects()
        self._register_task_targets_tests()
        self._register_extra_targets()
        self._gather_dependencies()
        self._build_target_graph()

    def _gather_dependencies(self) -> None:
        """Gather dependencies for each target"""
        logger.info("Gathering dependencies")

        # Collect a list of (build target, Python module) pairs from *all* the targets
        target_paths = []
        for target in self._targets.values():
            for dirpath, dirnames, filenames in os.walk(target.package_path):
                if os.path.basename(dirpath) in PROJECT_CONFIG.ignore_dirs:
                    # Empty the list of directories so os.walk does not recur
                    dirnames.clear()
                else:
                    for filename in filenames:
                        if filename.endswith(".py"):
                            target_paths.append(
                                (target, os.path.join(dirpath, filename))
                            )

        # Create a pool of workers that will parse imports from each Python module
        # path. Default worker count equals the number of CPU cores.
        paths = [path for _, path in target_paths]
        with Pool() as pool:
            # Return a list of sets containing imported package names
            imported_package_names_per_path = pool.map(
                gather_dependencies_from_module, paths
            )

        # Allow each target to set depencies in one batch. For most targets this will
        # be a no-op.
        for target in self._targets.values():
            target.set_dependencies(self._targets)

        # For each target path / package name pair, process and add the dependency to
        # the target's set of dependencies.
        for (target, _), package_names in zip(
            target_paths, imported_package_names_per_path
        ):
            for package_name in package_names:
                target.add_dependency(self._targets, package_name)

    def generate_build_files(self) -> None:
        """Generate BUILD files.

        You must have already called :py:meth:`.register_packages`.
        """
        logger.info("Generating BUILD files")
        trees_to_write = []
        for target in self._targets.values():
            tree = target.generate_build_file()
            if tree is not None:
                trees_to_write.append((tree, target.build_file))

        # Create a pool of workers that will render, format, and write each AST module
        # tree to disk. Default worker count equals the number of CPU cores.
        with Pool() as pool:
            pool.starmap(write_build_file, trees_to_write)

    def generate_build_file(self, target_key: str) -> None:
        """Generate BUILD file for a single target.

        You must have already called :py:meth:`.register_packages`.
        """
        target = self._targets[target_key]
        tree = target.generate_build_file()
        write_build_file(tree, target.build_file)

    def _register_task_targets_code(self) -> None:
        """Register task targets for code packages

        This iterates through various directories looking for a setup.py file. If it
        finds one, it means we've found a Python package and can register a build
        target.
        """
        logger.info("Registering code targets")

        setup_py_paths: List[Tuple[str, Path]] = []
        for top_dir_name in PROJECT_CONFIG.top_dirs:
            logger.debug(f"Walking {top_dir_name}")
            for dirpath, dirnames, filenames in os.walk(top_dir_name):
                if os.path.basename(dirpath) in PROJECT_CONFIG.ignore_dirs:
                    # Empty the list of directories so os.walk does not recur
                    dirnames.clear()
                else:
                    for filename in filenames:
                        if filename == "setup.py":
                            setup_py_paths.append(
                                (top_dir_name, Path(dirpath).joinpath(filename))
                            )

        for top_dir_name, setup_py_path in setup_py_paths:
            logger.debug(f"Registering {setup_py_path}")

            src_dir = setup_py_path.parent.joinpath("src")
            if not src_dir.is_dir():
                continue
            src_entries = [
                src_entry
                for src_entry in os.scandir(src_dir)
                if (
                    src_entry.is_dir()
                    and "egg-info" not in src_entry.path
                    and "pycache" not in src_entry.path
                )
            ]
            if len(src_entries) == 0:
                logger.debug(f"No entries found in {src_dir}")
                continue
            elif len(src_entries) > 1:
                raise MultipleSourcePackagesFound(
                    f"More than one package found in {src_dir}:"
                    f" {src_entries}. This could be caused by old git"
                    " state. Either manually clean up the packages or"
                    " run `git clean -fxd` if you are OK with a"
                    " completely clean slate."
                )

            src_entry = src_entries[0]
            package_name = os.path.basename(src_entry.path)
            config = Config(setup_py_path.parent)
            package_type = config.type
            if package_type is None:
                # Default the package type to "library"
                PythonPackageClass = PythonLibraryPackage
                config.set("package", "type", "library")
            else:
                PythonPackageClass = BUILD_TARGET_MAP[package_type]
            target = PythonPackageClass(
                target_type="code",
                build_template=top_dir_name,
                top_dir_name=top_dir_name,
                package_dir_name=str(setup_py_path.parent.relative_to(top_dir_name)),
                package_path=src_entry.path,
                package_name=package_name,
                build_dir=str(src_dir),
                config=config,
            )
            if target.key in PROJECT_CONFIG.ignore_targets:
                logger.debug(f"Ignoring {target}")
            elif package_name in self._targets:
                raise DuplicateTarget(
                    f"Duplicate target found for {package_name}. Check the"
                    " git repository for old folders that have the same"
                    " name."
                )
            else:
                self._targets[package_name] = target
                logger.debug(f"Registered target {target}")

    def _register_task_targets_tests(self) -> None:
        """Register task targets for tests

        This iterates through the previously registered build targets for code and
        registers test targets if they exist.
        """
        logger.info("Registering test targets")
        _targets = copy.copy(self._targets)
        for code_target in _targets.values():
            for tests_dir_type in ["unit", "functional", "component"]:
                tests_dir = os.path.join(
                    code_target.top_dir_name,
                    code_target.package_dir_name,
                    "tests",
                    tests_dir_type,
                )
                try:
                    os.scandir(tests_dir)
                except FileNotFoundError:
                    logger.debug(f"No tests found in {tests_dir}")
                else:
                    target = PythonTestPackage(
                        target_type="tests",
                        build_template="tests",
                        top_dir_name=code_target.top_dir_name,
                        package_dir_name=code_target.package_dir_name,
                        package_path=tests_dir,
                        package_name=tests_dir,
                        build_dir=tests_dir,
                        extra_tags={tests_dir_type},
                        code_target_package_name=code_target.package_name,
                    )
                    if target.key in PROJECT_CONFIG.ignore_targets:
                        logger.debug(f"Ignoring {target}")
                    else:
                        self._targets[tests_dir] = target
                        logger.debug(f"Registered target {target}")

    def _register_task_targets_py2sfn_projects(self) -> None:
        """Register py2sfn project targets.

        This looks for stepfunctions/projects/* folders. The BUILD file in each project
        consists of a generic ``target`` that only serves to point to the underlying
        tasks.
        """
        project_paths = Path("stepfunctions/projects").glob("*")
        for project_path in project_paths:
            target = PY2SFNProjectPackage(
                target_type="py2sfn-project",
                build_template="py2sfn-project",
                top_dir_name="stepfunctions/projects",
                package_dir_name=project_path.name,
                package_path=str(project_path),
                package_name=str(project_path),
                build_dir=str(project_path),
            )
            if target.key in PROJECT_CONFIG.ignore_targets:
                logger.debug(f"Ignoring {target}")
            elif target.key in self._targets:
                raise DuplicateTarget(
                    f"Duplicate target found for {target.key}. Check the"
                    " git repository for old folders that have the same"
                    " name."
                )
            else:
                self._targets[target.key] = target
                logger.debug(f"Registered target {target}")

    def _register_extra_targets(self) -> None:
        """Register extra project-specific targets.

        If the project has a ``.pypants/targets.py`` module defined with a top-level ``register_extra_targets``
        function defined, this method will call it and update the dictionary of build targets.
        """
        targets_path = PROJECT_CONFIG.config_dir_path.joinpath(".pypants/targets.py")
        if not targets_path.exists():
            logger.debug(f"No project targets.py found at {targets_path}")
            return

        logger.info("Registering extra project-specific targets")
        spec = importlib.util.spec_from_file_location("targets", str(targets_path))
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)
        targets = mod.register_extra_targets()
        existing_target_names = set(self._targets.keys()).intersection(
            set(targets.keys())
        )
        if len(existing_target_names) > 0:
            logger.warning(
                f"Extra target(s) already registered: {', '.join(existing_target_names)}"
            )
        self._targets.update(targets)

    def _build_target_graph(self) -> None:
        """Build a graph of package targets.

        This is used to check for cycles and compute dependencies.
        """
        G = nx.DiGraph()
        for target in self._targets.values():
            G.add_node(target)
            for dependency in target.dependencies:
                G.add_edge(target, dependency)
        self._target_graph = G

    def ensure_no_circular_imports(self) -> None:
        """Ensure there are no circular imports in any of the processed packages

        You must have already called :py:meth:`.register_packages`.

        Raises:
            :py:exec:`.CircularImportsFound` if any circular imports were found

        """
        logger.info("Ensuring no circular imports")
        cycles = list(nx.simple_cycles(self._target_graph))
        if len(cycles) > 0:
            raise CircularImportsFound(cycles)

    def compute_package_dependencies(
        self, target_key: str, include_3rdparty: bool = False
    ) -> List[PythonPackage]:
        """Compute a topologically-sorted list of dependencies for a given target.

        Args:
            target_key: Key of the registered build target
            include_3rdparty: Whether to include or exclude 3rdparty requirements in
                the list of dependencies

        Returns:
            list of PythonPackage instances sorted topologically

        """
        try:
            target = self._targets[target_key]
        except KeyError:
            raise NoTargetFound(
                f"No target registered named {target_key}."
                f" Available targets: {', '.join(sorted(self._targets.keys()))}"
            )

        dependencies = nx.descendants(self._target_graph, target)
        # Create a new graph with only the dependencies so we can sort
        subgraph = nx.DiGraph(self._target_graph.subgraph(dependencies))
        dependencies = list(nx.topological_sort(subgraph))
        if not include_3rdparty:
            dependencies = [
                d for d in dependencies if not isinstance(d, PythonRequirement)
            ]

        return dependencies



from typing import Dict

import click

from pypants.config import PROJECT_CONFIG
from pypants.generators.base import PythonPackageGenerator


class CLIPackageGenerator(PythonPackageGenerator):
    """Generator for a Python CLI package"""

    PACKAGE_TYPE = "cli"
    TOP_LEVEL_DIR = "packages"
    FOLDER_NAME_PREFIX = "cli_"

    @property
    def binary_name(self):
        """CLI binary name"""
        return f"example-{self.nickname}"

    @property
    def context(self) -> Dict:
        """Template context variables"""
        context = super().context
        context.update({"binary_name": self.binary_name, "platforms": "darwin,linux"})
        return context

    def print_extra_help(self) -> None:
        """Print extra help info to the terminal"""
        click.echo()
        click.secho("Try out your new CLI:", fg="green", bold=True)
        click.echo(
            f"{PROJECT_CONFIG.config_dir_path}/pants run {self.build_dir}:{self.binary_name} -- welcome"
        )



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['pypants', 'pypants.build_targets', 'pypants.generators']

package_data = \
{'': ['*']}

install_requires = \
['astor<1,>=0.7.1',
 'black==19.3b0',
 'Click>=7.0',
 'cookiecutter<2,>=1.6.0',
 'networkx<3,>=2.2',
 'python-slugify<2,>=1.2.0']

entry_points = \
{'console_scripts': ['pypants = pypants.cli:cli']}

setup(name='pypants',
      version='0.1.4',
      description='CLI for working with Python packages and BUILD files in a Pants monorepo',
      author='Jonathan Drake',
      author_email='jdrake@narrativescience.com',
      url='https://github.com/NarrativeScience/pypants',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      entry_points=entry_points,
      python_requires='>=3.6,<4',
     )



"""Contains a configuration parser and project config singleton"""
from configparser import ConfigParser
import json
from pathlib import Path
from typing import List, Optional, Set

from .util import get_git_top_level_path


class Config:
    """pypants configuration parser

    Configuration is based on a .pants.cfg file.

    The schema for the file in the git project root is::

        [project]
        ignore_dirs = []
        ignore_targets = []
        python_package_name_prefix = ""
        third_party_import_map_path = "3rdparty/python/import-map.json"
        third_party_requirements_path = "3rdparty/python/requirements.txt"
        top_dirs = ["."]

    The schema for the file in an individual python package is::

        [package]
        extra_dependencies = ["lib/python_core/src", ...]
        extra_tags = ["my-tag", ...]
        generate_build_file = <bool>
        generate_pytest_binary = <bool>
        include_test_coverage = <bool>
        type = "library"|"binary"|...

    Instantiating this class will load and initialize a config parser instance.
    """

    def __init__(self, config_dir_path: str) -> None:
        """

        Args:
            config_dir_path: path to a folder containing .pants.cfg file

        """
        self.config_dir_path = config_dir_path
        self._pants_cfg_path = Path(config_dir_path).joinpath(".pants.cfg")
        self._config = ConfigParser()
        self._config["project"] = {}
        self._config["package"] = {}
        if self._pants_cfg_path.is_file():
            self._config.read(self._pants_cfg_path)

    def set(self, *args) -> None:
        """Proxy method to set a value on the underlying config parser instance"""
        self._config.set(*args)

    # Project options

    @property
    def ignore_dirs(self) -> Set[str]:
        """Never look for or process files in these directories"""
        return frozenset(
            json.loads(self._config.get("project", "ignore_dirs", fallback="[]"))
        )

    @property
    def ignore_targets(self) -> Set[str]:
        """Set of target package names to ignore when collecting build targets"""
        return frozenset(
            json.loads(self._config.get("project", "ignore_targets", fallback="[]"))
        )

    @property
    def python_package_name_prefix(self) -> str:
        """Prefix to use for names of packages generated by pypants"""
        return self._config.get("project", "python_package_name_prefix", fallback="")

    @property
    def third_party_import_map_path(self) -> Path:
        """Path to the location of the import-map.json file relative to the project root"""
        return Path(
            self._config.get(
                "project",
                "third_party_import_map_path",
                fallback="3rdparty/python/import-map.json",
            )
        )

    @property
    def third_party_requirements_path(self) -> Path:
        """Path to the requirements.txt relative to the project root"""
        return Path(
            self._config.get(
                "project",
                "third_party_requirements_path",
                fallback="3rdparty/python/requirements.txt",
            )
        )

    @property
    def top_dirs(self) -> List[str]:
        """Top-level directories to search for Python packages"""
        return json.loads(self._config.get("project", "top_dirs", fallback="['.']"))

    # Package options

    @property
    def extra_dependencies(self) -> Set[str]:
        """Extra set of dependencies to include in the python_library target"""
        return set(
            json.loads(self._config.get("package", "extra_dependencies", fallback="[]"))
        )

    @property
    def extra_tags(self) -> Set[str]:
        """Extra set of tags to include in the Pants build targets"""
        return set(json.loads(self._config.get("package", "extra_tags", fallback="[]")))

    @property
    def generate_build_file(self) -> bool:
        """Flag denoting whether to generate a BUILD file"""
        return self._config.getboolean("package", "generate_build_file", fallback=True)

    @property
    def generate_pytest_binary(self) -> bool:
        """Flag denoting whether to include a python_binary target for pytest"""
        return self._config.getboolean(
            "package", "generate_pytest_binary", fallback=False
        )

    @property
    def include_test_coverage(self) -> bool:
        """Flag denoting whether to include a coverage attribute on pytest targets"""
        return self._config.getboolean(
            "package", "include_test_coverage", fallback=True
        )

    @property
    def type(self) -> Optional[str]:
        """Package type.

        Returning none means we should use a heuristic to figure out the type.
        """
        return self._config.get("package", "type", fallback=None)


# Create a singleton for the project configuration
PROJECT_CONFIG = Config(get_git_top_level_path())



"""Contains PythonBinaryPackage class"""
import ast
import logging
import os
from typing import Tuple

import astor

from ..exceptions import NoConsoleScriptFound  # noqa
from .python_package import PythonPackage  # noqa

logger = logging.getLogger(__name__)


class GetConsoleScript(ast.NodeVisitor):
    """AST node visitor for finding the console_script string in setup.py"""

    def __init__(self) -> None:
        super(GetConsoleScript).__init__()
        # Example specifier: ``mybinary = my_package.cli:main``
        self.console_script_specifier = None

    def visit_keyword(self, node) -> None:
        """Visitor for keyword nodes"""
        if node.arg == "entry_points":
            for key, value in zip(node.value.keys, node.value.values):
                if key.s == "console_scripts":
                    self.console_script_specifier = value.elts[0].s
                    break


class PythonBinaryPackage(PythonPackage):
    """Represents a Python CLI or server package build target in Pants

    CLIs and servers have an entry point, meaning they are meant to be executed
    directly.
    """

    @property
    def dependency_target(self) -> str:
        """Returns the representation of this target in another target's dependencies

        It's not good practice to depend on a server/binary package (apart from
        tests), but if that happens we need to set the dependency target to
        `:lib` so it correctly references the `python_library` target.
        """
        return f"{self.key}:lib"

    def _parse_entry_point(self) -> Tuple[str, str]:
        """Get the CLI binary name and source module path from setup.py

        Returns:
            Tuple of:
                * **binary name**
                * **source module path**

        """
        setup_py_path = os.path.join(self.package_dir_path, "setup.py")
        tree = astor.code_to_ast.parse_file(setup_py_path)
        visitor = GetConsoleScript()
        visitor.visit(tree)
        if visitor.console_script_specifier is None:
            raise NoConsoleScriptFound(
                f"Could not generate BUILD file for {self.key} because no"
                f" console_scripts found in {setup_py_path}"
            )

        logger.debug(
            f"Found console_script in {setup_py_path}:"
            f" {visitor.console_script_specifier}"
        )
        # Example specifier: ``mybinary = my_package.cli:main``
        # Final binary name: ``mybinary``
        # Final source module path: ``my_package/cli.py``
        binary_name, entry_point = visitor.console_script_specifier.split("=")
        binary_name = binary_name.strip()
        entry_point_parts = entry_point.strip().split(":")[0].split(".")
        entry_point_parts[-1] += ".py"
        source_module_path = "/".join(entry_point_parts)

        logger.debug(
            f"Using binary_name={binary_name} source_module_path={source_module_path}"
        )
        return binary_name, source_module_path

    def _generate_python_binary_cli_ast_node(self) -> ast.Expr:
        """Generate an AST node for a python_binary Pants target"""
        binary_name, source_module_path = self._parse_entry_point()
        node = ast.Expr(
            value=ast.Call(
                func=ast.Name(id="python_binary"),
                args=[],
                keywords=[
                    ast.keyword(arg="name", value=ast.Str(binary_name)),
                    ast.keyword(
                        arg="dependencies", value=ast.List(elts=[ast.Str(":lib")])
                    ),
                    ast.keyword(arg="source", value=ast.Str(source_module_path)),
                    self._tags_keyword,
                ],
            )
        )
        return node

    def generate_build_file_ast_node(self) -> ast.Module:
        """Generate a Pants BUILD file as an AST module node"""
        node = ast.Module(
            body=[
                self._generate_python_library_ast_node(
                    name="lib", globs_path=f"{self.package_name}/**/*"
                ),
                self._generate_python_binary_cli_ast_node(),
            ]
        )
        return node
